  .inesprg 1   ; 1x 16KB PRG code
  .ineschr 1   ; 1x  8KB CHR data
  .inesmap 0   ; mapper 0 = NROM, no bank swapping
  .inesmir 1   ; background mirroring
  

;;;;;;;;;;;;;;;

room_index  .rs 1           ;which room to load
ptr1  .rs 2                 ;pointer to the background information

data_y  .rs 1               ;y counter
data_x  .rs 1               ;x counter

repeat_meta_tile  .rs 1     ;times to repeat the current meta tile

meta_tile_sets_ptr  .rs 2   ;pointer to the meta tile sets
meta_tile_ptr  .rs 2        ;finding the meta tile that is active

background_row  .rs $20     ;the second row that each meta tile contains
 
joypad1  .rs 1              ;these 3 handle the controllers
joypad1_old  .rs 1
joypad1_pressed  .rs 1

sleeping  .rs 1             ;run main program once per frame
updating_background  .rs 1  ;disable the NMI updates while updating the background


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    
  .bank 0
  .org $C000 

;;;;;;;;;;;;;;

RESET:
  SEI          ; disable IRQs
  CLD          ; disable decimal mode
  LDX #$40
  STX $4017    ; disable APU frame IRQ
  LDX #$FF
  TXS          ; Set up stack
  INX          ; now X = 0
  STX $2000    ; disable NMI
  STX $2001    ; disable rendering
  STX $4010    ; disable DMC IRQs

vblankwait1:       ; First wait for vblank to make sure PPU is ready
  BIT $2002
  BPL vblankwait1

clrmem:
  LDA #$00
  STA $0000, x
  STA $0100, x
  STA $0300, x
  STA $0400, x
  STA $0500, x
  STA $0600, x
  STA $0700, x
  LDA #$FE
  STA $0200, x
  INX
  BNE clrmem
   
vblankwait2:      ; Second wait for vblank, PPU is ready after this
  BIT $2002
  BPL vblankwait2

LoadPalettes:
  LDA $2002             ; read PPU status to reset the high/low latch
  LDA #$3F
  STA $2006             ; write the high byte of $3F00 address
  LDA #$00
  STA $2006             ; write the low byte of $3F00 address
  LDX #$00              ; start out at 0
LoadPalettesLoop:
  LDA palette, x        
  STA $2007             ; write to PPU
  INX                   ; X = X + 1
  CPX #$10              ; Compare X to hex $10, decimal 16 - copying 16 bytes = 4 sprites
  BNE LoadPalettesLoop  ; Branch to LoadPalettesLoop if compare was Not Equal to zero
                        ; if compare was equal to 32, keep going down
                           
LoadAttribute:
  LDA $2002             ; read PPU status to reset the high/low latch
  LDA #$23
  STA $2006             ; write the high byte of $23C0 address
  LDA #$C0
  STA $2006             ; write the low byte of $23C0 address
  LDX #$00 
  LDA #%11111111
LoadAttributeLoop:
  STA $2007
  INX
  CPX #$40
  BNE LoadAttributeLoop

  LDA #$00
  STA room_index

  JSR LoadbgBackground

  LDA #$00
  STA $2003       ; set the low byte (00) of the RAM address             
              
  LDA #%10010000   ; disable NMI, sprites from Pattern Table 0, background from Pattern Table 1
  STA $2000

  LDA #%00001110   ; disable sprites, enable background, no clipping on left side
  STA $2001

;----------------------------------------------------------------------
;-----------------------START MAIN PROGRAM-----------------------------
;----------------------------------------------------------------------

Forever:
  
  INC sleeping

.loop
  LDA sleeping
  BNE .loop

  JSR strobe_controllers

  JSR handle_input

  JMP Forever     ;jump back to Forever, infinite loop
  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 
;--------THE FUCKIN' NMI ROUTINE, RECOGNIZE BITCH!------------;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

NMI:

  LDA updating_background
  BNE skip_graphics_updates

  LDA #$00        ;no scrolling
  STA $2005
  STA $2005

  LDA #%00001110  ;disable sprites,enable background,no clipping
  STA $2001

WakeUp:
  LDA #$00
  STA sleeping

skip_graphics_updates

  RTI             ; return from interrupt

;;;;;;;;;;;;;; 

;this section handles the input from the controllers and runs stuff

handle_input:

ReadUp:
  LDA joypad1_pressed	;Up
  AND #%00001000
  BEQ ReadUpDone
  
  LDA #$00
  STA room_index

  JMP LoadbgBackground

ReadUpDone:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

ReadDown:
  LDA joypad1_pressed	;Down
  AND #%00000100
  BEQ ReadDownDone

  LDA #$02
  STA room_index

  JMP LoadbgBackground

ReadDownDone:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

ReadRight:
  LDA joypad1_pressed	;Right
  AND #%00000001
  BEQ ReadRightDone

  LDA #$04
  STA room_index

  JMP LoadbgBackground

ReadRightDone:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

ReadLeft:
  LDA joypad1_pressed	;Left
  AND #%00000010
  BEQ ReadLeftDone
  
  LDA #$06
  STA room_index

  JMP LoadbgBackground

ReadLeftDone:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

  RTS

;-----------------------------------

; read_joypad will capture the current button state and store it in joypad1.  
;       Off-to-on transitions will be stored in joypad1_pressed
strobe_controllers:
    lda joypad1
    sta joypad1_old ;save last frame's joypad button states
    
    lda #$01
    sta $4016
    lda #$00
    sta $4016
    
    ldx #$08
.loop:    
    lda $4016
    lsr a
    rol joypad1  ;A, B, select, start, up, down, left, right
    dex
    bne .loop
    
    lda joypad1_old ;what was pressed last frame.  EOR to flip all the bits to find ...
    eor #$FF    ;what was not pressed last frame
    and joypad1 ;what is pressed this frame
    sta joypad1_pressed ;stores off-to-on transitions
    
    rts

;-------------------------------

backgroundpointer:
  .word RoomUp,RoomDown,RoomRight,RoomLeft

;-------------------------------


LoadbgBackground:

  LDA #$01
  STA updating_background       ;"disable" NMI

  LDA #%00000110   ; disable sprites, disable background, no clipping on left side
  STA $2001

  LDX room_index
  LDA backgroundpointer,X
  STA ptr1
  LDA backgroundpointer+1,X
  STA ptr1+1

  LDA $2002             ; read PPU status to reset the high/low latch
  LDA #$20
  STA $2006             ; write the high byte of $2000 address
  LDA #$00
  STA $2006             ; write the low byte of $2000 address
  LDY #$00              ; start out at 00
  STY data_y
  LDX #$00
  STX data_x

background_start:

  LDA repeat_meta_tile         ;see if you are repeating a tile
  BEQ .jump 
  DEC repeat_meta_tile         ;decrement the repeated tile and jump back to the decompression routine
  JMP decompress

.jump

  LDY data_y
  LDA [ptr1],y
  CMP #$FF           ;see if it is at the end of the data string
  BNE .next
  JMP background_complete
.next
  LDA [ptr1],y
  AND #%10000000        ;see if this is a repeated tile or a switch of meta tile banks
  BNE .next1
  JMP .loadtile

.next1
  LDA [ptr1],y
  AND #%01000000        ;see if this is a switch of the meta tile banks
  BNE .next2
  JMP .repeat

.next2
  LDA [ptr1],y
  AND #%00111111         ;switch meta tile banks
  ASL A
  TAX

  LDA meta_tile_sets,x   ;load the new meta tile bank
  STA meta_tile_sets_ptr
  LDA meta_tile_sets+1,x
  STA meta_tile_sets_ptr+1
  INY
  STY data_y 
  JMP background_start      ;jump to the next entry in the data table and return to the start of the routine

.repeat
  LDA [ptr1],y
  AND #%00111111        ;load number of repeats for this meta tile 
  STA repeat_meta_tile  ;note that this is set up so that you need to have the number of repeats - 1
  INY
  STY data_y

.loadtile
  LDA [ptr1],y        ;load the meta tile number
  INY
  STY data_y
  ASL A               ;this is a table of addresses, so multiply by 2
  TAY

  LDA [meta_tile_sets_ptr],y    ;load the address of the meta tile in question
  STA meta_tile_ptr
  INY
  LDA [meta_tile_sets_ptr],y
  STA meta_tile_ptr+1
  
decompress:
  LDY #$00
  LDX data_x
  LDA [meta_tile_ptr],y      ;write the first tile of the meta tile to 2007
  STA $2007
  INY
  LDA [meta_tile_ptr],y      ;write the second tile of the meta tile to 2007
  STA $2007
  INY
  LDA [meta_tile_ptr],y      ;write the third tile of the meta tile to memory for later use
  STA background_row,x
  INY
  INX
  LDA [meta_tile_ptr],y      ;write the forth tile of the meta tile to memory for later use
  STA background_row,x
  INX
  STX data_x
  CPX #$20                   ;see if we are at the end of a row.  If so, we need to use the tiles stored in memory, if not, jump to the top.
  BEQ .next
  JMP background_start
  
.next
  LDY #$00
row_start:                  ;write the tiles stored in memory to the background
  LDA background_row,y
  STA $2007
  INY
  CPY #$20
  BNE row_start
  LDX #$00
  STX data_x

  JMP background_start      ;jump back to the start of the routine and continue on

background_complete:

  LDA #$00                  ;"enable" NMI
  STA updating_background

  RTS

;-----------------------------------------------

;-------------------------------------------------

palette:
  .db $22,$29,$1A,$0F,  $22,$36,$17,$0F,  $22,$30,$21,$0F,  $22,$27,$17,$0F   ;;background palette

;;;;;;;;;;;;;;  
  
  .bank 1
  .org $E000

;;;;;;;;;;;;;;;


RoomUp:
  .db %11000000,$8F,$06
  .db $06,$03,$03,$82,$04,$83,$03,$82,$04,$03,$03,$06
  .db $8F,$06
  .db $06,$8D,$04,$06
  .db $8F,$06
  .db $06,$00,$06,$00,$06,$00,$06,$00,$06,$00,$06,$00,$06,$00,$06,$06
  .db $06,$8D,$02,$06
  .db $06,$03,$03,$82,$04,$83,$03,$82,$04,$03,$03,$06
  .db $8F,$06
  .db $06,$8D,$04,$06
  .db $8F,$06
  .db $8F,$06
  .db $06,$05,$06,$05,$06,$06,$05,$06,$06,$05,$06,$06,$05,$06,$05,$06
  .db %11000001,$8F,$00
  .db $8F,$01,$FF

;------------------------------------------------

RoomDown:
  .db %11000000,$06,$01,$01,$06,$01,$01,$06,$01,$01,$06,$01,$01,$06,$01,$01,$06
  .db $8F,$06
  .db $82,$06,$02,$06,$02,$06,$02,$06,$02,$06,$02,$83,$06
  .db $8F,$06
  .db $8F,$06
  .db $06,$06,%11000001,$04,$89,$05,$06,%11000000,$06,$06
  .db $86,$06,%11000001,$01,$01,%11000000,$86,$06
  .db $06,$83,$04,$06,$06,%11000001,$01,$01,%11000000,$06,$06,$83,$04,$06
  .db $06,$83,$04,$06,$06,%11000001,$01,$01,%11000000,$06,$06,$83,$04,$06
  .db $86,$06,%11000001,$01,$01,%11000000,$86,$06
  .db $86,$06,%11000001,$01,$01,%11000000,$86,$06
  .db $06,%11000001,$03,%11000000,$06,%11000001,$03,%11000000,$06,%11000001,$03,%11000000,$06,%11000001,$01,$01,%11000000,$06,%11000001,$03,%11000000,$06,%11000001,$03,%11000000,$06,%11000001,$03,%11000000,$06
  .db $06,%11000001,$02,%11000000,$06,%11000001,$02,%11000000,$06,%11000001,$02,%11000000,$06,%11000001,$01,$01,%11000000,$06,%11000001,$02,%11000000,$06,%11000001,$02,%11000000,$06,%11000001,$02,%11000000,$06
  .db %11000001,$8F,$00
  .db $8F,$01,$FF

;------------------------------------------------

RoomRight:
  .db %11000000,$8F,$06
  .db $06,$83,$03,$06,$83,$03,$06,$83,$03,$06
  .db $8F,$06
  .db $82,$06,%11000001,$03,%11000000,$06,%11000001,$03,%11000000,$06,%11000001,$03,$03,%11000000,$06,%11000001,$03,%11000000,$06,%11000001,$03,%11000000,$82,$06
  .db $82,$06,%11000001,$02,%11000000,$06,%11000001,$02,%11000000,$06,%11000001,$02,$02,%11000000,$06,%11000001,$02,%11000000,$06,%11000001,$02,%11000000,$82,$06
  .db $82,$06,%11000001,$89,$00,%11000000,$82,$06
  .db $82,$06,%11000001,$89,$01,%11000000,$82,$06
  .db $86,$06,%11000001,$01,$01,%11000000,$86,$06
  .db $06,$00,$00,$06,$02,$02,$06,%11000001,$01,$01,%11000000,$06,$02,$02,$06,$00,$00,$06
  .db $06,$00,$00,$06,$02,$02,$06,%11000001,$01,$01,%11000000,$06,$02,$02,$06,$00,$00,$06
  .db $86,$06,%11000001,$01,$01,%11000000,$86,$06
  .db $86,$06,%11000001,$01,$01,%11000000,$86,$06
  .db $86,$06,%11000001,$01,$01,%11000000,$86,$06
  .db %11000001,$8F,$00
  .db $8F,$01,$FF

;-----------------------------------------------

RoomLeft:
  .db %11000000,$8F,$06
  .db $06,$00,$00,$83,$04,$00,$00,$83,$04,$00,$00,$06
  .db $06,$00,$00,$03,$03,$00,$00,$03,$03,$00,$00,$03,$03,$00,$00,$06
  .db $8F,$06
  .db $83,$06,$05,$05,$06,$05,$05,$06,$05,$05,$83,$06
  .db $82,$06,%11000001,$04,$87,$05,$06,%11000000,$82,$06
  .db $86,$06,%11000001,$01,$01,%11000000,$86,$06
  .db $06,$06,%11000001,$03,%11000000,$06,%11000001,$03,%11000000,$06,$06,%11000001,$01,$01,%11000000,$06,$06,%11000001,$03,%11000000,$06,%11000001,$03,%11000000,$06,$06
  .db $06,$06,%11000001,$02,%11000000,$06,%11000001,$02,%11000000,$06,$06,%11000001,$01,$01,%11000000,$06,$06,%11000001,$02,%11000000,$06,%11000001,$02,%11000000,$06,$06
  .db $06,%11000001,$8D,$00,%11000000,$06
  .db $06,%11000001,$8D,$01,%11000000,$06
  .db $06,$06,$82,$02,$06,$06,%11000001,$01,$01,%11000000,$06,$06,$82,$02,$06,$06
  .db $06,$02,$82,$06,$02,$06,%11000001,$01,$01,%11000000,$06,$02,$82,$06,$02,$06
  .db %11000001,$8F,$00
  .db $8F,$01,$FF

;-------------

meta_tile_sets:
  .word MetaTileSeta,MetaTileSetb

;-------------

MetaTileSeta:
  .word MetaTileSeta00,MetaTileSeta01,MetaTileSeta02,MetaTileSeta03,MetaTileSeta04,MetaTileSeta05,MetaTileSeta06

;-------------

MetaTileSetb:
  .word MetaTileSetb00,MetaTileSetb01,MetaTileSetb02,MetaTileSetb03,MetaTileSetb04,MetaTileSetb05,MetaTileSetb06

;--------------

MetaTileSeta00
  .db $44,$49,$5F,$7A

MetaTileSeta01
  .db $AB,$AD,$AC,$AE

MetaTileSeta02
  .db $53,$54,$55,$56

MetaTileSeta03
  .db $B0,$B2,$B1,$B3

MetaTileSeta04
  .db $A5,$A6,$A7,$A8

MetaTileSeta05
  .db $7B,$7C,$7D,$7E

MetaTileSeta06
  .db $24,$24,$24,$24

;--------------------------

MetaTileSetb00
  .db $45,$45,$47,$47

MetaTileSetb01
  .db $47,$47,$47,$47

MetaTileSetb02
  .db $CA,$CB,$CC,$CD

MetaTileSetb03
  .db $C6,$C7,$C8,$C9

MetaTileSetb04
  .db $24,$6B,$24,$70

MetaTileSetb05
  .db $6C,$6D,$71,$72

MetaTileSetb06
  .db $6F,$24,$74,$24

;--------------------------





;;;;;;;;;;;;;;;;;;;;;;;;;;

  .org $FFFA     ;first of the three vectors starts here
  .dw NMI        ;when an NMI happens (once per frame if enabled) the 
                   ;processor will jump to the label NMI:
  .dw RESET      ;when the processor first turns on or is reset, it will jump
                   ;to the label RESET:
  .dw 0          ;external interrupt IRQ is not used in this tutorial
  
  
;;;;;;;;;;;;;;  
  
  
  .bank 2
  .org $0000
  .incbin "mario.chr"   ;includes 8KB graphics file from SMB1
